import { Api } from '../../Api'
import { BaseWorkingGroupFixture } from './BaseWorkingGroupFixture'
import { QueryNodeApi } from '../../QueryNodeApi'
import { EventDetails, WorkingGroupModuleName } from '../../types'
import { SubmittableExtrinsic } from '@polkadot/api/types'
import { Utils } from '../../utils'
import { ISubmittableResult } from '@polkadot/types/types/'
import { StatusTextChangedEventFieldsFragment } from '../../graphql/generated/queries'
import { EventType } from '../../graphql/generated/schema'
import { assert } from 'chai'
import { RemoveUpcomingOpening, WorkingGroupMetadataAction } from '@joystream/metadata-protobuf'

export class RemoveUpcomingOpeningsFixture extends BaseWorkingGroupFixture {
  protected upcomingOpeningIds: string[]

  public constructor(api: Api, query: QueryNodeApi, group: WorkingGroupModuleName, upcomingOpeningIds: string[]) {
    super(api, query, group)
    this.upcomingOpeningIds = upcomingOpeningIds
  }

  protected async getSignerAccountOrAccounts(): Promise<string> {
    return this.api.getLeadRoleKey(this.group)
  }

  protected async getExtrinsics(): Promise<SubmittableExtrinsic<'promise'>[]> {
    return this.upcomingOpeningIds.map((id) => {
      const metaBytes = Utils.metadataToBytes(this.getActionMetadata(id))
      return this.api.tx[this.group].setStatusText(metaBytes)
    })
  }

  protected async getEventFromResult(result: ISubmittableResult): Promise<EventDetails> {
    return this.api.retrieveWorkingGroupsEventDetails(result, this.group, 'StatusTextChanged')
  }

  protected getActionMetadata(upcomingOpeningId: string): WorkingGroupMetadataAction {
    const actionMeta = new WorkingGroupMetadataAction()
    const removeUpcomingOpeningMeta = new RemoveUpcomingOpening()
    removeUpcomingOpeningMeta.setId(upcomingOpeningId)
    actionMeta.setRemoveUpcomingOpening(removeUpcomingOpeningMeta)

    return actionMeta
  }

  protected assertQueryNodeEventIsValid(qEvent: StatusTextChangedEventFieldsFragment, i: number): void {
    assert.equal(qEvent.event.type, EventType.StatusTextChanged)
    assert.equal(qEvent.group.name, this.group)
    assert.equal(qEvent.metadata, Utils.metadataToBytes(this.getActionMetadata(this.upcomingOpeningIds[i])).toString())
    Utils.assert(qEvent.result.__typename === 'UpcomingOpeningRemoved', 'Unexpected StatuxTextChangedEvent result type')
    assert.equal(qEvent.result.upcomingOpeningId, this.upcomingOpeningIds[i])
  }

  async runQueryNodeChecks(): Promise<void> {
    await super.runQueryNodeChecks()
    // Query & check the event
    await this.query.tryQueryWithTimeout(
      () => this.query.getStatusTextChangedEvents(this.events),
      (qEvents) => this.assertQueryNodeEventsAreValid(qEvents)
    )
    // Query the openings and make sure they no longer exist
    await Promise.all(
      this.upcomingOpeningIds.map(async (id) => {
        const qUpcomingOpening = await this.query.getUpcomingOpeningById(id)
        assert.isNull(qUpcomingOpening)
      })
    )
  }
}
